@@ -1,7 +1,8 @@
 #! /usr/bin/env python
 __authors__ = ["Dilmurat Yusuf", "Bert Droesbeke"]
 __copyright__ = "Copyright 2020, Dilmurat Yusuf"
-__email__ = "Dilmurat.yusuf@gmail.com"
+__maintainer__ = "Bert Droesbeke"
+__email__ = "bedro@psb.vib-ugent.be"
 __license__ = "MIT"
 
 import os
@@ -20,10 +21,19 @@ import pandas as pd
 import tempfile
 from ena_upload._version import __version__
 
+SCHEMA_TYPES = ['study', 'experiment', 'run', 'sample']
 
-# SettingWithCopyWarning causes false positive
-# e.g at df.loc[:, 'file_checksum'] = md5
-pd.options.mode.chained_assignment = None
+
+class MyFTP_TLS(ftplib.FTP_TLS):
+    """Explicit FTPS, with shared TLS session"""
+
+    def ntransfercmd(self, cmd, rest=None):
+        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)
+        if self._prot_p:
+            conn = self.context.wrap_socket(conn,
+                                            server_hostname=self.host,
+                                            session=self.sock.session)
+        return conn, size
 
 
 def create_dataframe(schema_tables, action):
@@ -39,27 +49,12 @@ def create_dataframe(schema_tables, action):
                           dataframe -- pandas dataframe
     '''
 
-    # ? would it be good to use alias to index rows?
-
     schema_dataframe = {}
 
     for schema, table in schema_tables.items():
-        df = pd.read_csv(table, sep='\t', comment='#', dtype = str)
-        # checking for optional columns and if not present adding them
-        if schema == 'sample':
-            optional_columns = ['accession', 'submission_date', 'status', 'scientific_name', 'taxon_id']
-        else:
-            optional_columns = ['accession', 'submission_date', 'status']
-        for header in optional_columns:
-            if not header in df.columns:
-                if header == 'status':
-                    df[header] = action.lower()
-                else:
-                    df[header] = np.nan
-        # status column contain action keywords
-        # for xml rendering, keywords require uppercase
-        # according to scheme definition of submission
-        df['status'] = df['status'].str.upper()
+        df = pd.read_csv(table, sep='\t', comment='#', dtype=str)
+        df = df.dropna(how='all')
+        df = check_columns(df, schema, action)
         schema_dataframe[schema] = df
 
     return schema_dataframe
@@ -85,6 +80,29 @@ def extract_targets(action, schema_dataframe):
     return schema_targets
 
 
+def check_columns(df, schema, action):
+    # checking for optional columns and if not present, adding them
+    if schema == 'sample':
+        optional_columns = ['accession', 'submission_date',
+                            'status', 'scientific_name', 'taxon_id']
+    elif schema == 'run':
+        optional_columns = ['accession',
+                            'submission_date', 'status', 'file_checksum']
+    else:
+        optional_columns = ['accession', 'submission_date', 'status']
+
+    for header in optional_columns:
+        if not header in df.columns:
+            if header == 'status':
+                # status column contain action keywords
+                # for xml rendering, keywords require uppercase
+                # according to scheme definition of submission
+                df[header] = str(action).upper()
+            else:
+                df[header] = np.nan
+
+    return df
+
 def check_filenames(file_paths, run_df):
     """Compare data filenames from command line and from RUN table.
 
@@ -150,7 +168,10 @@ def generate_stream(schema, targets, Template, center, tool):
     if schema == 'run':
         # These attributes are required for rendering
         # the run xml templates
-        file_attrib = ['file_name', 'file_format', 'file_checksum']
+        # Adding backwards compatibility for file_format
+        if 'file_format' in targets:
+            targets.rename(columns={'file_format': 'file_type'}, inplace=True)
+        file_attrib = ['file_name', 'file_type', 'file_checksum']
         other_attrib = ['alias', 'experiment_alias']
         run_groups = targets[other_attrib].groupby(targets['alias'])
         run_groups = run_groups.experiment_alias.unique()
@@ -261,6 +282,8 @@ def construct_submission(template_path, action, submission_input, center, checkl
     :return submission_xml: filename of submission XML
     '''
 
+    print("Constructing submission")
+
     xsds, templates = actors(template_path, checklist)
 
     template = templates['submission']
@@ -297,6 +320,7 @@ def get_md5(filepath):
 
     return md5sum.hexdigest()
 
+
 def get_taxon_id(scientific_name):
     """Get taxon ID for input scientific_name.
 
@@ -318,6 +342,7 @@ def get_taxon_id(scientific_name):
         msg = f'Oops, no taxon ID avaible for {scientific_name}. Is it a valid scientific name?'
         sys.exit(msg)
 
+
 def get_scientific_name(taxon_id):
     """Get scientific name for input taxon_id.
 
@@ -343,22 +368,32 @@ def submit_data(file_paths, password, webin_id):
     :param file_paths: a dictionary of filename string and file_path string
     :param args: the command-line arguments parsed by ArgumentParser
     """
+    ftp_host = "webin2.ebi.ac.uk"
 
+    print("\nConnecting to ftp.webin2.ebi.ac.uk....")
     try:
-        print("\nConnecting to ftp.webin.ebi.ac.uk....")
-        ftp = ftplib.FTP("webin.ebi.ac.uk", webin_id, password)
-    except IOError:
-        print(ftp.lastErrorText())
+        ftps = MyFTP_TLS(timeout=10)
+        ftps.context.set_ciphers('HIGH:!DH:!aNULL')
+        ftps.connect(ftp_host, port=21)
+        ftps.auth()
+        ftps.login(webin_id, password)
+        ftps.prot_p()
+
+    except IOError as ioe:
+        print(ioe)
         print("ERROR: could not connect to the ftp server.\
                Please check your login details.")
-
+        sys.exit()
     for filename, path in file_paths.items():
         print(f'uploading {path}')
-        ftp.storbinary(f'STOR {filename}', open(path, 'rb'))
-        msg = ftp.storbinary(f'STOR {filename}', open(path, 'rb'))
-        print(msg)
+        try:
+            print(ftps.storbinary(f'STOR {filename}', open(path, 'rb')))
+        except BaseException as err:
+            print(f"ERROR: {err}")
+            print("ERROR: If your connection times out at this stage, it propably is because of a firewall that is in place. FTP is used in passive mode and connection will be opened to one of the ports: 40000 and 50000.")
+            raise
+    print(ftps.quit())
 
-    print(ftp.quit())
 
 def columns_to_update(df):
     '''
@@ -415,7 +450,7 @@ def process_receipt(receipt, action):
     success = receipt_root.get('success')
 
     if success == 'true':
-        print('\nSubmission was done successfully')
+        print('Submission was done successfully')
     else:
         errors = []
         for element in receipt_root.findall('MESSAGES/ERROR'):
@@ -442,28 +477,44 @@ def process_receipt(receipt, action):
         return df
 
     receiptDate = receipt_root.get('receiptDate')
-
-    study_update = receipt_root.findall('STUDY')
-    sample_update = receipt_root.findall('SAMPLE')
-    experiment_update = receipt_root.findall('EXPERIMENT')
-    run_update = receipt_root.findall('RUN')
-
     schema_update = {}  # schema as key, dataframe as value
+    if action in ['ADD', 'MODIFY']:
+        study_update = receipt_root.findall('STUDY')
+        sample_update = receipt_root.findall('SAMPLE')
+        experiment_update = receipt_root.findall('EXPERIMENT')
+        run_update = receipt_root.findall('RUN')
 
-    if study_update:
-        schema_update['study'] = make_update(study_update, 'study')
+        if study_update:
+            schema_update['study'] = make_update(study_update, 'study')
 
-    if sample_update:
-        schema_update['sample'] = make_update(sample_update, 'sample')
+        if sample_update:
+            schema_update['sample'] = make_update(sample_update, 'sample')
 
-    if experiment_update:
-        schema_update['experiment'] = make_update(
-            experiment_update, 'experiment')
+        if experiment_update:
+            schema_update['experiment'] = make_update(
+                experiment_update, 'experiment')
 
-    if run_update:
-        schema_update['run'] = make_update(run_update, 'run')
+        if run_update:
+            schema_update['run'] = make_update(run_update, 'run')
+        return schema_update
 
-    return schema_update
+    # release does have the accession numbers that are released in the recipe
+    elif action == 'RELEASE':
+        receipt_info = {}
+        infoblocks = receipt_root.findall('MESSAGES/INFO')
+        for element in infoblocks:
+            match = re.search('(.+?) accession "(.+?)"', element.text)
+            if match and match.group(1) in receipt_info:
+                receipt_info[match.group(1)].append(match.group(2))
+            elif match and match.group(1) not in receipt_info:
+                receipt_info[match.group(1)] = [match.group(2)]
+        for ena_type, accessions in receipt_info.items():
+            print(f"\n{ena_type.capitalize()} accession details:")
+            update_list = []
+            for accession in accessions:
+                extract = (accession, receiptDate, status[action])
+                update_list.append(extract)
+                print("\t".join(extract))
 
 
 def update_table(schema_dataframe, schema_targets, schema_update):
@@ -505,20 +556,51 @@ def update_table(schema_dataframe, schema_targets, schema_update):
                 dataframe.loc[index,
                               'taxon_id'] = targets.loc[index, 'taxon_id']
             elif schema == 'run':
-                # since index is set to alias
+                # Since index is set to alias
                 # then targets of run can have multiple rows with
                 # identical index because each row has a file
-                # which is associated with a run.
+                # which is associated with a run
                 # and a run can have multiple files.
-                # the following assignment assumes 'targets' retain
-                # the orginal row order in 'dataframe'
-                # because targets was initially subset of 'datafram'.
+                # The following assignment assumes 'targets' retain
+                # the original row order in 'dataframe'
+                # because targets was initially subset of 'dataframe'.
                 dataframe.loc[index,
                               'file_checksum'] = targets.loc[index, 'file_checksum']
 
     return schema_dataframe
 
 
+def update_table_simple(schema_dataframe, schema_targets, action):
+    """Update schema_dataframe with info in schema_targets.
+
+    :param schema_dataframe: a dictionary - {schema:dataframe}
+    :param_targets: a dictionary - {schema:targets}
+
+    'schema' -- a string - 'study', 'sample','run', 'experiment'
+    'dataframe' -- a pandas dataframe created from the input tables
+    'targets' -- a filtered dataframe with 'action' keywords
+                 contains updated columns - md5sum and taxon_id
+
+    :return schema_dataframe: a dictionary - {schema:dataframe}
+                              dataframe -- updated status
+    """
+    # define expected status based on action
+    status = {'ADD': 'added', 'MODIFY': 'modified',
+              'CANCEL': 'cancelled', 'RELEASE': 'released'}
+
+    for schema in schema_dataframe:
+        dataframe = schema_dataframe[schema]
+        targets = schema_targets[schema]
+
+        dataframe.set_index('alias', inplace=True)
+        targets.set_index('alias', inplace=True)
+
+        for index in targets.index:
+            dataframe.loc[index, 'status'] = status[action]
+
+    return schema_dataframe
+
+
 def save_update(schema_tables_, schema_dataframe_):
     """Write updated dataframe to tsv file.
 
@@ -529,7 +611,7 @@ def save_update(schema_tables_, schema_dataframe_):
         :dataframe: a dataframe
     """
 
-    print('\nSaving updates in new tsv tables::')
+    print('\nSaving updates in new tsv tables:')
     for schema in schema_tables_:
         table = schema_tables_[schema]
         dataframe = schema_dataframe_[schema]
@@ -537,7 +619,7 @@ def save_update(schema_tables_, schema_dataframe_):
         file_name, file_extension = os.path.splitext(table)
         update_name = f'{file_name}_updated{file_extension}'
         dataframe.to_csv(update_name, sep='\t')
-        print(f'save updates in {update_name}')
+        print(f'{update_name}')
 
 
 class SmartFormatter(argparse.HelpFormatter):
@@ -602,7 +684,10 @@ def process_args():
 
     parser.add_argument('--checklist', help="specify the sample checklist with following pattern: ERC0000XX, Default: ERC000011", dest='checklist',
                         default='ERC000011')
-
+    
+    parser.add_argument('--xlsx',
+                        help='excel table with metadata')
+    
     parser.add_argument('--tool',
                         dest='tool_name',
                         default='ena-upload-cli',
@@ -634,7 +719,7 @@ def process_args():
 
     # check if any table is given
     tables = set([args.study, args.sample, args.experiment, args.run])
-    if tables == {None}:
+    if tables == {None} and not args.xlsx:
         parser.error('Requires at least one table for submission')
 
     # check if .secret file exists
@@ -642,9 +727,15 @@ def process_args():
         if not os.path.isfile(args.secret):
             msg = f"Oops, the file {args.secret} does not exist"
             parser.error(msg)
+    
+    # check if xlsx file exists
+    if args.xlsx:
+        if not os.path.isfile(args.xlsx):
+            msg = f"Oops, the file {args.xlsx} does not exist"
+            parser.error(msg)
 
     # check if data is given when adding a 'run' table
-    if args.action == 'add' and args.run is not None:
+    if (not args.no_data_upload and args.run) or (not args.no_data_upload and args.xlsx):
         if args.data is None:
             parser.error('Oops, requires data for submitting RUN object')
 
@@ -673,6 +764,16 @@ def collect_tables(args):
 
     return schema_tables
 
+def update_date(date):
+    if pd.isnull(date) or isinstance(date, str):
+        return date
+    try:
+        return date.strftime('%Y-%m-%d')
+    except AttributeError:
+        return date
+    except Exception:
+        raise
+
 
 def main():
     args = process_args()
@@ -683,6 +784,7 @@ def main():
     checklist = args.checklist
     secret = args.secret
     draft = args.draft
+    xlsx = args.xlsx
 
     with open(secret, 'r') as secret_file:
         credentials = yaml.load(secret_file, Loader=yaml.FullLoader)
@@ -695,13 +797,32 @@ def main():
             f"Oops, file {args.secret} does not contain a password or username")
     secret_file.close()
 
-    # ? a function needed to convert characters e.g. # -> %23 in password
+    if xlsx:
+        # create dataframe from xlsx table
+        xl_workbook = pd.ExcelFile(xlsx)
+        schema_dataframe = {}  # load the parsed data in a dict: sheet_name -> pandas_frame
+        schema_tables = {}
 
-    # collect the schema with table input from command-line
-    schema_tables = collect_tables(args)
+        for schema in SCHEMA_TYPES:
+            xl_sheet = xl_workbook.parse(schema, header=0)
+            xl_sheet = xl_sheet.drop(0).dropna(how='all')
+            for column_name in list(xl_sheet.columns.values):
+                if 'date' in column_name:
+                    xl_sheet[column_name] = xl_sheet[column_name].apply(update_date)
 
-    # create dataframe from table
-    schema_dataframe = create_dataframe(schema_tables, action)
+            if True in xl_sheet.columns.duplicated():
+                sys.exit("Duplicated columns found")
+
+            xl_sheet = check_columns(xl_sheet, schema, action)
+            schema_dataframe[schema] = xl_sheet
+            path = os.path.dirname(os.path.abspath(xlsx))
+            schema_tables[schema] = f"{path}/ENA_template_{schema}.tsv"
+    else:
+        # collect the schema with table input from command-line
+        schema_tables = collect_tables(args)
+
+        # create dataframe from table
+        schema_dataframe = create_dataframe(schema_tables, action)
 
     # ? add a function to sanitize characters
     # ? print 'validate table for specific action'
@@ -712,7 +833,8 @@ def main():
     schema_targets = extract_targets(action, schema_dataframe)
 
     if not schema_targets:
-        sys.exit(f"There is no table submitted having at least one row with {action} as action in the status column.")
+        sys.exit(
+            f"There is no table submitted having at least one row with {action} as action in the status column.")
 
     if action == 'ADD':
         # when adding run object
@@ -721,34 +843,43 @@ def main():
         if 'run' in schema_targets:
             # a dictionary of filename:file_path
             df = schema_targets['run']
-               
-            file_paths = {os.path.basename(path): os.path.abspath(path)
-                                                for path in args.data}
-            # check if file names identical between command line and table
-            # if not, system exits
-            check_filenames(file_paths, df)
-
+            file_paths = {}
+            if args.data:
+                for path in args.data:
+                    file_paths[os.path.basename(path)] =  os.path.abspath(path) 
+                # check if file names identical between command line and table
+                # if not, system exits
+                check_filenames(file_paths, df)
+            
             # generate MD5 sum if not supplied in table
-            if not check_file_checksum(df):
+            if file_paths and not check_file_checksum(df):
                 print("No valid checksums found, generate now...", end=" ")
                 file_md5 = {filename: get_md5(path) for filename, path
                             in file_paths.items()}
 
                 # update schema_targets wih md5 hash
                 md5 = df['file_name'].apply(lambda x: file_md5[x]).values
+                # SettingWithCopyWarning causes false positive
+                # e.g at df.loc[:, 'file_checksum'] = md5
+                pd.options.mode.chained_assignment = None
                 df.loc[:, 'file_checksum'] = md5
                 print("done.")
+            elif check_file_checksum(df):
+                print("Valid checksums found", end=" ")
+            else:
+                sys.exit("No valid checksums found and no files given to generate checksum from. Please list the files using the --data option or specify the checksums in the run-table when the data is uploaded separately.")
 
             schema_targets['run'] = df
 
             # submit data to webin ftp server
             if args.no_data_upload:
-                print("No files will be uploaded, remove `--no_data_upload' argument to perform upload.")
+                print(
+                    "No files will be uploaded, remove `--no_data_upload' argument to perform upload.")
             elif draft:
-                print("No files will be uploaded, remove `--draft' argument to perform upload.")
+                print(
+                    "No files will be uploaded, remove `--draft' argument to perform upload.")
             else:
                 submit_data(file_paths, password, webin_id)
-                  
 
         # when adding sample
         # update schema_targets with taxon ids or scientific names
@@ -765,7 +896,8 @@ def main():
                     scientificName = get_scientific_name(row['taxon_id'])
                     df.loc[index, 'scientific_name'] = scientificName
                 elif pd.isna(row['taxon_id']) and pd.isna(row['scientific_name']):
-                    sys.exit(f"No taxon_id or scientific_name was given with sample {row['alias']}.")
+                    sys.exit(
+                        f"No taxon_id or scientific_name was given with sample {row['alias']}.")
             print('Taxon IDs and scientific names are retrieved')
             schema_targets['sample'] = df
 
@@ -790,6 +922,8 @@ def main():
         submission_xml = construct_submission(template_path, action,
                                               schema_targets, center, checklist, tool)
 
+    else:
+        print(f"The action {action} is not supported.")
     schema_xmls['submission'] = submission_xml
 
     if draft:
@@ -811,14 +945,19 @@ def main():
             print("There was an ERROR during submission:")
             sys.exit(receipt)
 
-        schema_dataframe = update_table(schema_dataframe,
-                                        schema_targets,
-                                        schema_update)
-
-        # save updates in new tables
-        save_update(schema_tables, schema_dataframe)
+        if action in ['ADD', 'MODIFY']:
+            schema_dataframe = update_table(schema_dataframe,
+                                            schema_targets,
+                                            schema_update)
+            # save updates in new tables
+            save_update(schema_tables, schema_dataframe)
+        elif action in ['CANCEL', 'RELEASE']:
+            schema_dataframe = update_table_simple(schema_dataframe,
+                                                   schema_targets,
+                                                   action)
+            # save updates in new tables
+            save_update(schema_tables, schema_dataframe)
 
 
 if __name__ == "__main__":
-
     main()
